#ifndef WIN32
#include <dlfcn.h>
#else
#include <windows.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include "LibLoader.h"
#include "StringBuilder.h"
#include "icommands.h"
#include "functions.h"

#ifndef bool
#define bool char
#define true 1
#define false 0
#endif

#ifndef New
#define New(x) (struct x *)malloc(sizeof(struct x))
#endif

#ifndef P
#define P(x) struct x *
#endif

#ifndef Del
#define Del(x) {free(x); x = NULL;}
#endif

#ifndef DelN
#define DelN(x) {if (x != NULL){free(x); x = NULL;}}
#endif

#ifdef __cplusplus
extern "C" {
#endif

void *library;
char *lib_path;
void *function;
const char delimeter = 0;
int needRun;

/**
 * Constructor of library
 */
static void __attribute__((constructor)) init() {
	needRun = true;
	library = NULL;
}

/**
 * Destructor of library
 */
static void __attribute__((destructor)) finit() {
	if (library != NULL) {
		unloadLibrary();
	}
	if (lib_path != NULL) {
		free(lib_path);
		lib_path = NULL;
	}
}

/**
 * Load library by path
 * @param s path to library (or name)
 * @return true, if library loaded successfull, false else
 */
bool loadLibrary(const char * const s) {
	lib_path = strdup(s);
#ifndef WIN32
	library = dlopen(lib_path, RTLD_LAZY);
	if (!library) {
		fprintf(stderr, "dlopen() error: %s\n", dlerror());
                fflush(stderr);
		return false;
	}
#else
	library = LoadLibrary(lib_path);
	if (!library) {
		fprintf(stderr, "Cannot open the library %s\n", lib_path);
                fflush(stderr);
		return false;
	}
#endif
	return true;
}

/**
 * Load function from current library
 * @param f name of function to load
 * @return true if library loaded successfull, false else
 */
bool loadFunction(const char * const f) {
#ifdef DEBUG
	FILE *out = fopen("errors.txt", "a+");
	fprintf(out, "loadFunction(%s) started\n", f);
	fclose(out);
#endif
	if (library == NULL) {
		if (lib_path != NULL) {
#ifndef WIN32
			library = dlopen(lib_path, RTLD_LAZY);
#else
			library = LoadLibrary(lib_path);
#endif
			if (!library) {
				return false;
			}
		}
		return false;
	}
#ifndef WIN32
	function = dlsym(library, f);
#else
	function = GetProcAddress((HINSTANCE)library, f);
#endif
	if (!function) {
		char * s = dlerror();

#ifdef DEBUG
		out = fopen("errors.txt", "a+");
		fprintf(out, "dlsym() error: %s\n", s);
		fprintf(out, "Error: cannot load function %s in the library %s\n", f, lib_path);
		fclose(out);
#endif
#ifndef WIN32
		fprintf(stderr, "dlsym() error: %s\n", s);
#else
		fprintf(stderr, "Error: cannot load function %s in the library %s\n", f, lib_path);
#endif
		fflush(stderr);
#ifdef DEBUG
		out = fopen("errors.txt", "a+");
		fprintf(out, "And return false\n");
		fclose(out);
#endif
		return false;
	}
#ifdef DEBUG
	out = fopen("errors.txt", "a+");
	fprintf(out, "function %s loaded, address: %p\n", f, function);
	fclose(out);
#endif
	return true;
}

/**
 * Unload current library
 */
void unloadLibrary() {
	if (library == NULL) {
		return;
	}
#ifndef WIN32
	int rt = dlclose(library);
	if (rt != 0) {
		fprintf(stderr, "dlclose() error: %s\n", dlerror());
		fflush(stderr);
	}
#else
	FreeLibrary((HINSTANCE)library);
#endif
	library = NULL;
	if (lib_path != NULL) {
		free(lib_path);
		lib_path = NULL;
	}
	return;
}

/**
 * Is library loaded?
 * @return true, if library loaded, false else
 */
bool isLoadedLibrary() {
	return (library != NULL);
}

/**
 * Return next input string before delimeter
 * @return pointer to C string, if EOF reached, return NULL
 */
char * getNextString() {
	P(StringBuilder) sb = sbNew();
	char ch;
	int r;
	while (1) {
		r = fgetc(stdin);
		if (r == EOF) {
#ifdef DEBUG
			FILE *out = fopen("errors.txt", "a+");
			char *s;
			fprintf(out, "Next string (break): %s\n", s = sbToString(sb));
			sbPrintOut(out, sb);
			Del(s);
			fclose(out);
#endif
			sbFinit(&sb);
			needRun = false;
			return NULL;
		}
		ch = (char) r;
		if (ch == delimeter) {
			break;
		}
		sbAddChar(sb, ch);
	}
	char *res = sbToString(sb);
	sbFinit(&sb);
#ifdef DEBUG
	FILE *out = fopen("errors.txt", "a+");
	fprintf(out, "Input string: %s\n", res);
	fclose(out);
#endif
	return res;
}

/**
 * Treat loadLibrary command
 * @param s to compare with lLoad constant
 * @return 0, if s is not command to load library, non-zero else
 */
int treatLoad(const char * const s) {
	if (!strcmp(s, lLoadC)) {
		char *l = getNextString();
		if (l == NULL) {
			needRun = false;
			return 2;
		}
		int r = 1;
		if (!loadLibrary(l)) {
			r = 2;
		}
		Del(l);
		return r;
	}
	return 0;
}
/**
 * Treat loadLibrary command
 * @param s to compare with lLoad constant
 * @return 0, if s is not command to load library, non-zero else
 */
int treatUnload(const char * const s) {
	if (!strcmp(s, uLoadC)) {
#ifdef DEBUG
			FILE *out = fopen("errors.txt", "a+");
			fprintf(out, "Called unloadLibrary\n");
			fclose(out);
#endif
		unloadLibrary();
		return 1;
	}
	return 0;
}
/**
 * Check string for run function command
 * @param s string to check
 * @param res result of calling function
 * @return non-zero, if s is "run" string (1 if run successfull, 2 else), 0 if s is not "run" string
 */
int treatRun(const char * const s, char **res) {
	if (!strcmp(s, runC)) {
		char *f = getNextString();
		if (f == NULL) {
			needRun = 0;
			return 2;
		}
		if (!loadFunction(f)) {
			(*res) = strdup(failedA);
#ifdef DEBUG
			FILE *out = fopen("errors.txt", "a+");
			fprintf(out, "Answer: %s\n", (*res));
			fclose(out);
#endif
			Del(f);
			return 2;
		}
		if (!strcmp(f, runF)) {
			void (*r)(void) = (void(*)(void)) function;
#ifdef DEBUG
			FILE *out = fopen("errors.txt", "a+");
			fprintf(out, "Function run'll start now\n");
			fclose(out);
#endif
			r();
#ifdef DEBUG
			out = fopen("errors.txt", "a+");
			fprintf(out, "Function is finished now\n");
			fclose(out);
#endif
			(*res) = strdup(goodA);
		}
		if (!strcmp(f, getNameF)) {
			const char * const (*r)(void) =
					(const char * const (*)(void)) function;
			(*res) = strdup(r());
		}
		if (!strcmp(f, "treatEvent")) {
			void (*r)(const char * const ) =
					(void(*)(const char * const )) function;
			r(getNextString());
			(*res) = strdup(goodA);
		}
		Del(f);
		return 1;
	}
	return 0;
}

/**
 * Check string for query, if string is "query",
 * get query's id and query itself
 * @param s string to check
 * @param queryName text of query, if s is "query" string
 * @param id id of query, if s is "query" string
 * @return 1, if s is query, 0 else
 */
int treatQuery(const char * const s, char **queryName, char **id) {
	if (!strcmp(s, queryS)) {
		(*id) = getNextString();
		(*queryName) = getNextString();
		return 1;
	}
	return 0;
}

/**
 * Check string for answer on query, if string is "answer", 
 * get answer's id and answer itself
 * @param s string to check
 * @param ans text of answer, if s is "answer" string
 * @param id id of answer, if s is "answer" string
 * @return 1, if s is answer, 0 else
 */
int treatAnswer(const char* const s, char ** ans, char **id) {
	if (!strcmp(s, answerS)) {
		(*id) = getNextString();
		(*ans) = getNextString();
		return 1;
	}
	return 0;
}

/**
 * Exec query
 */
char * execQuery(const char * const query) {
	char * res = NULL;
	if (treatRun(query, &res)) {
		return res;
	}
	int r;
	r = treatLoad(query);
	if (r) {
		if (r == 2) {
			return strdup(failedA);
		}
		return strdup(goodA);
	}
	if (treatUnload(query)) {
		return strdup(goodA);
	}
	return res;
}

/**
 * Treat input stream
 */
void treatInput() {
	char *s;
	int count_zero = 0;
	char *query, *id, *res;
	s = query = id = res = NULL;
	while (needRun) {
		s = getNextString();
		if (s == NULL) {
			needRun = false;
#ifdef DEBUG
			FILE *out = fopen("errors.txt", "a+");
			fprintf(out, "Next string is null (%s)\n", s);
			fclose(out);
#endif
			continue;
		}
		if (s[0] == '\0') {
			count_zero++;
		} else {
			count_zero = 0;
		}
		if (count_zero >= 2) {
			needRun = false;
#ifdef DEBUG
			FILE *out = fopen("errors.txt", "a+");
			fprintf(out, "Count of zero =  %d\n", count_zero);
			fclose(out);
#endif
			continue;
		}
		if (s[0] == '\0') {
			Del(s);
			continue;
		}
		if (!treatQuery(s, &query, &id)) {
			fprintf(stderr, "Unknown command (%s)\n", s);
			fflush(stderr);
#ifdef DEBUG
			FILE *out = fopen("errors.txt", "a+");
			fprintf(out, "Unknown command (%s)\n", s);
			fclose(out);
#endif
			Del(s);
			continue;
		}
		res = execQuery(query);
		fprintf(stdout, "%s%c%s%c%s%c", answerS, delimeter, id, delimeter, res,
				delimeter);
		fflush(stdout);
#ifdef DEBUG
		FILE *out = fopen("errors.txt", "a+");
		fprintf(out, "%s%c%s%c%s%c\n", answerS, delimeter, id, delimeter, res, delimeter);
		fclose(out);
#endif
		Del(s);
		Del(query);
		Del(id);
		Del(res);
	}
	DelN(s);
	DelN(query);
	DelN(id);
	DelN(res);
}

#ifdef __cplusplus
}
#endif
