#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "StringBuilder.h"

#ifndef New
#define New(x) (struct x *)malloc(sizeof(struct x))
#endif

#ifndef P
#define P(x) struct x *
#endif

#ifndef Min
#define Min(x, y) ((x < y)? x : y)
#endif

/**
 * Constructor of StringBuilder
 * @param p pointer to StringBuilder
 */
void sbInit(P(StringBuilder) p) {
	p->last_string = 0;
	p->character_pos = 0;
	p->array = (char **) malloc(sizeof(char*) * (p->last_string + 1));
	if (p->array == NULL) {
		perror("Not enough memory\n");
		exit(2);
	}
	p->array[p->last_string] = (char*) malloc(sizeof(char) * StringBuilder_MAX);
	if (p->array[p->last_string] == NULL) {
		perror("Not enough memory\n");
		exit(2);
	}
}

/**
 * Check for empty StringBuilder
 * @return non-zero value, if StringBuilder is empty, 0 else
 */
int sbIsEmpty(P(StringBuilder) p) {
	return (p->last_string == 0 && p->character_pos == 0);
}

/**
 * Create new instance of StringBuilder
 * @return pointer to new instance of StringBuilder
 */
P(StringBuilder) sbNew() {
	P(StringBuilder) res = New(StringBuilder);
	sbInit(res);
	return res;
}

/**
 * Destructor of StringBuilder
 * @param p pointer to pointer to StringBuilder
 */
void sbFinit(P(StringBuilder) *p) {
	int i;
	if ((*p)->array == NULL) {
		return;
	}
	for (i = 0; i <= (*p)->last_string; i++) {
		if ((*p)->array[i] != NULL) {
			free((*p)->array[i]);
			(*p)->array[i] = NULL;
		}
	}
	free((*p)->array);
	(*p)->array = NULL;
	free(*p);
}

/**
 * Get string from StringBuilder
 * @param p pointer to StringBuilder
 */
char * sbToString(P(StringBuilder) p) {
	if (p == NULL) {
		fprintf(stderr, "Null pointer error, file %s, line %d\n", __FILE__,
				__LINE__);
		exit(3);
	}
	if (p->array == NULL) {
		return NULL;
	}
	int n = p->last_string * StringBuilder_MAX + p->character_pos + 1;
	char *res = (char *) malloc(n * sizeof(char));
	int pos, i, j;
	pos = 0;
	for (i = 0; i < p->last_string; i++) {
		for (j = 0; j < StringBuilder_MAX; j++) {
			res[pos++] = p->array[i][j];
		}
	}
	for (j = 0; j < p->character_pos; j++) {
		res[pos++] = p->array[p->last_string][j];
	}
	res[pos] = '\0';
	return res;
}
/**
 * Print content of StringBuilder for debug purpose
 */
void sbPrint(P(StringBuilder) p) {
	int i, j;
	fprintf(stderr, "StringBuilder starts\n");
	for (i = 0; i < p->last_string; i++) {
		for (j = 0; j < StringBuilder_MAX; j++) {
			fprintf(stderr, "%c", p->array[i][j]);
		}
		fprintf(stderr, "\n");
	}
	fprintf(stderr, "%s\n", p->array[p->last_string]);
	fprintf(stderr, "StringBuilder ends\n");
	fflush(stderr);
}

/**
 * Print content of StringBuilder for debug purpose to out
 */
void sbPrintOut(P(StringBuilder) p, FILE * out) {
	int i, j;
	fprintf(out, "StringBuilder starts\n");
	for (i = 0; i < p->last_string; i++) {
		for (j = 0; j < StringBuilder_MAX; j++) {
			fprintf(out, "%c", p->array[i][j]);
		}
		fprintf(out, "\n");
	}
	fprintf(out, "%s\n", p->array[p->last_string]);
	fprintf(out, "StringBuilder ends\n");
	fflush(out);
}

/**
 * Increase size of StringBuilder for StringBuilder_MAX characters
 * and change StringBuilder->character_pos to 0.
 * @note private function
 * @param p pointer to StringBuilder
 */
void sbIncreaseSize(P(StringBuilder) p) {
	char ** res = (char**) realloc(p->array, sizeof(char *) * (p->last_string
			+ 2));
	if (res == NULL) {
		perror("Not enough memory\n");
		exit(2);
	}
	p->array = res;
	p->last_string++;
	p->array[p->last_string]
			= (char *) malloc(sizeof(char) * StringBuilder_MAX);
	p->character_pos = 0;
	p->array[p->last_string][p->character_pos] = '\0';
}

/**
 * Add character to StringBuilder
 * @param ch character to add
 * @param p pointer to StringBuilder
 */
void sbAddChar(P(StringBuilder) p, const char ch) {
	if (p == NULL) {
		fprintf(stderr, "Null pointer error, file %s, line %d\n", __FILE__,
				__LINE__);
		fflush(stderr);
		exit(3);
	}
	p->array[p->last_string][p->character_pos++] = ch;
	if (p->character_pos >= StringBuilder_MAX) {
		sbIncreaseSize(p);
	}
	p->array[p->last_string][p->character_pos] = '\0';
}

/**
 * Add string to StringBuilder
 * @param s string to add
 * @param p pointer to StringBuilder
 */
void sbAddString(P(StringBuilder) p, const char * const s) {
	if (p == NULL) {
		fprintf(stderr, "Null pointer error, file %s, line %d\n", __FILE__,
				__LINE__);
		fflush(stderr);
		exit(3);
	}
	int ns = strlen(s);
	int i, m;
	int shift = 0;
	while (ns) {
		m = Min(ns, StringBuilder_MAX - p->character_pos);
		for (i = 0; i < m; i++) {
			p->array[p->last_string][p->character_pos + i] = s[shift + i];
		}
		p->character_pos += m;
		if (p->character_pos >= StringBuilder_MAX) {
			sbIncreaseSize(p);
		}
		ns -= m;
		shift += m;
	}
	p->array[p->last_string][p->character_pos] = '\0';
}
